; 
;
;               HASH LIBRARY                               by Nmeth Gbor
;
;
; 

    .486                       ; create 32 bit code
    .model flat, stdcall       ; 32 bit memory model
    option casemap :none       ; case sensitive

    include     \masm32\include\windows.inc
    include     \masm32\include\kernel32.inc
    include     \masm32\macros\macros.asm

; needed for debugging
    include     \masm32\include\masm32.inc
    include     \masm32\include\debug.inc

    includelib  \masm32\lib\kernel32.lib

; needed for debugging
    includelib  \masm32\lib\debug.lib
    includelib  \masm32\lib\masm32.lib

; 
;
; FUNCTION PROTOTYPES
;
; HASH1_init	PROTO :DWORD,:DWORD,:DWORD
; HASH1_destroy	PROTO :DWORD
; HASH1_store	PROTO :DWORD
; HASH1_search	PROTO :DWORD
;
; 
;
; HASH1 STRUCTURE for uniform 32bit size records
;

HASH1_STRUCT    STRUCT

                START     dd ?        ; start of hash array
                COUNT     dd ?        ; max entry number
                FREE      dd ?        ; number of free slots  
                GETKEY    dd ?        ; pointer of hashing function
                PROBE     dd ?        ; pointer of probe function
                MAXPROBE  dd ?        ; maximum number of probes
                
HASH1_STRUCT    ENDS

; 

.data
; needed for debugging

        PUBLIC collision
            collision   dd 0        ; number of collisions during storing

; 

.code
align 4
; 
;
; DEFAULT HASH FUNCTION
;
; When the user does not assign a hash function in HASH1_init this is used
;
; IN:   eax - input data
;       edx - address of hash structure
;
; OUT:  ebx - key/index
;
; -------------------------------------------------------------------------
HASH1_DEF_FUNC1 PROC USES eax ecx edx

                ASSUME edx:PTR HASH1_STRUCT
                mov     ebx,[edx].COUNT         ; get hash modulus (size)
                ASSUME edx:NOTHING

; this is a slightly modified version of rand proc in nrand.asm
    xor edx, edx
    mov ecx, 127773
    div ecx
    mov ecx, eax
    mov eax, 16807
    mul edx
    mov edx, ecx
    mov ecx, eax
    mov eax, 2836
    mul edx
    sub ecx, eax
    xor edx, edx
    mov eax, ecx

    div ebx
    mov ebx,edx                                 ; output is a random number

;PrintHex ebx,'key1'
                ret

HASH1_DEF_FUNC1 ENDP
; 
;
; DEFAULT PROBE FUNCTION (linear probe)
;
; When the user does not assign a probe function in HASH1_init this is used
;
; IN:   ebx - input key/index
;       edx - address of hash structure
;
; OUT:  ebx - key/index
;
; -------------------------------------------------------------------------
HASH1_DEF_FUNC2 PROC USES eax edx

                or      ebx,ebx
                jnz  @_1
                ASSUME edx:PTR HASH1_STRUCT
                mov     ebx,[edx].COUNT
                ASSUME edx:NOTHING
@_1:
                sub     ebx,1
;PrintHex ebx,'key2'

; needed for debugging
                add dword ptr collision,1           ; increase the number of collisions
                ret

HASH1_DEF_FUNC2 ENDP

; 
;
; Initialize HASH utility
;
; IN:   hd - address of structure for hash data
;       n - max. number of entries (modulus/size)
;       hn - hash function address
;
; OUT:  fail    eax - NULL
;       succes  eax - address of hash store area
;
; -------------------------------------------------------------------------
HASH1_init      PROC USES ebx esi hd:DWORD,n:DWORD,hf1:DWORD,hf2:DWORD

                mov     esi,hd
                ASSUME esi:PTR HASH1_STRUCT
                mov     eax,hf1
                mov     ebx,hf2
                cmp     eax,NULL
                jnz  @_1
                lea     eax,HASH1_DEF_FUNC1
                lea     ebx,HASH1_DEF_FUNC2
@_1:
                mov     [esi].GETKEY,eax
                mov     DWORD PTR [esi].MAXPROBE,0
                mov     [esi].PROBE,ebx
                mov     eax,n
                mov     [esi].COUNT,eax
                mov     [esi].FREE,eax
                shl     eax,2
                invoke GlobalAlloc,GPTR,eax
                mov     [esi].START,eax
                ASSUME esi:NOTHING
                ret

HASH1_init      ENDP

; 
;
; Destroys HASH utility
;
; IN: hd - pointer to a HASH structure
;
; -------------------------------------------------------------------------
HASH1_destroy   PROC USES esi hd:DWORD

                mov     esi,hd
                ASSUME esi:PTR HASH1_STRUCT
                xor     eax,eax
                mov     [esi].COUNT,eax
                mov     [esi].GETKEY,eax
                invoke HeapFree,FUNC(GetProcessHeap),0,[esi].START
                mov     [esi].START,eax
                ASSUME esi:NOTHING
                ret

HASH1_destroy   ENDP

; 

; 
;
; Stores a data in HASH
;
; IN:   hd - pointer to a HASH structure
;      eax - data to store
;
; OUT: ebx - key/index
;
; -------------------------------------------------------------------------
HASH1_store     PROC USES ecx edx esi hd:DWORD

                mov     edx,hd
                ASSUME edx:PTR HASH1_STRUCT
                xor     ecx,ecx                         ; init number of probes with 0
                mov     esi,[edx].START

                call    DWORD PTR [edx].GETKEY          ; hash function: ebx - key/index
@_1:
                cmp     DWORD PTR [esi+ebx*4],0         ; is the slot empty?
                jz @_2

                add     ecx,1                           ; increase number of probes
                call    DWORD PTR [edx].PROBE           ; hash function2 >> ebx: key/index
                jmp @_1
@_2:
                mov     [esi+ebx*4],eax                 ; store value
                cmp     ecx,[edx].MAXPROBE              ; current number of probes greater?
                jc  @_3
                mov     [edx].MAXPROBE,ecx              ; store new number of probes
@_3:
                sub     [edx].FREE,1
                ASSUME edx:NOTHING
@_4:
                ret

HASH1_store     ENDP

; 

; 
;
; Searches for data in HASH
;
; IN: hd - pointer to a HASH structure
;    eax - key of data to look for
;
; -------------------------------------------------------------------------
HASH1_search    PROC USES ecx edi hd:DWORD

                mov     edx,hd
                ASSUME edx:PTR HASH1_STRUCT
                mov     ecx,[edx].MAXPROBE
                mov     esi,[edx].START
                call    DWORD PTR [edx].GETKEY          ; hash function: ebx - key/index
@_2:
                mov     edi,[esi+ebx*4]                 ; read data
                or      edi,edi                         ; is empty?
                jz @_3
                cmp     edi,eax                         ; key found?
                jz @_4
                call    DWORD PTR [edx].PROBE           ; hash function: ebx - key/index
;PrintHex ecx,'probe'
                sub     ecx,1                           ; number of probes
                jnc  @_2
                ASSUME edx:NOTHING
; ----------------
; Empty...
@_3:
                xor     esi,esi
                ret
; ----------------
; Found...
@_4:
                shl     ebx,2
                add     esi,ebx
                ret

HASH1_search    ENDP

; 


end